/* .vitepress/theme/custom.css */

/* Import IBM Plex fonts */
@import url('https://fonts.googleapis.com/css2?family=IBM+Plex+Mono:ital,wght@0,400;0,500;0,600;0,700;1,400;1,500;1,600;1,700&family=IBM+Plex+Sans:ital,wght@0,400;0,500;0,600;0,700;1,400;1,500;1,600;1,700&display=swap');

:root {
  /* IBM Plex font families */
  --vp-font-family-base: 'IBM Plex Sans', sans-serif;
  --vp-font-family-mono: 'IBM Plex Mono', monospace;
  --vp-c-brand-1: #d95f5f;
  --vp-c-brand-2: #e18080;
  --vp-c-brand-3: #d95f5f;
}

.VPHero .VPImage {
    max-width: 760px;
    position: relative;
    border-radius: 8px;
}

.VPHero .image-container {
    padding: 32px;
}

/* do the following only for mobile */
@media (max-width: 768px) {
    .VPHero .container {
        flex-direction: column-reverse;
    }

    .VPHero .image {
        margin-top: 0;
    }
}
